/***********************************************************************\
*                          __   ______  _____ __                        *
*                   __ _  / /__<  / _ )/ __(_) /__ ___                  *
*                  /  ' \/  '_// / _  / _// / / -_|_-<                  *
*                 /_/_/_/_/\_\/_/____/_/ /_/_/\__/___/                  *
*                                                                       *
*  -------------------------------------------------------------------  *
*  This file is a part of mk1BFiles                                     *
*                                                                       *
*  mk1BFiles is free software; you can redistribute it and/or           *
*  modify it under the terms of the GNU General Public License          *
*  as published by the Free Software Foundation; either version 3       *
*  of the License, or (at your option) any later version.               *
*                                                                       *
*  This program is distributed in the hope that it will be useful,      *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of       *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
*  GNU General Public License for more details.                         *
*                                                                       *
*  You should have received a copy of the GNU General Public License    *
*  along with this program; if not, write to the Free Software          *
*  Foundation, Inc., 51 Franklin Street, Fifth Floor,                   *
*  Boston, MA  02110-1301, USA.                                         *
*                                                                       *
*  -------------------------------------------------------------------  *
*  Copyright (C) 2010, Clercin guillaume <clercin.guillaume@gmail.com>  *
*  Last modified: Fri, 17 Sep 2010 16:28:34 +0200                       *
\***********************************************************************/

#include <fcntl.h>
#include <getopt.h>
#include <malloc.h>
#include <math.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "display.h"
#include "timer.h"
#include "util.h"

#define NBFILES 100000

static unsigned long long int nbFiles = NBFILES;
static unsigned short width;
static unsigned long long iFiles = 0;
static unsigned short bshift = 10;
static pthread_mutex_t lock;
static pthread_mutex_t lSize;
static unsigned long long totalSize = 0;
static char * buffer = 0;
static int openOption = O_WRONLY | O_CREAT | O_TRUNC | O_DIRECT;

__attribute__ ((constructor))
void initMain() {
	pthread_mutex_init(&lock, 0);
	pthread_mutex_init(&lSize, 0);

	buffer = memalign(4096, 10240);
	bzero(buffer, 10240);
}

unsigned long long getIndex() {
	pthread_mutex_lock(&lock);
	unsigned long long index = iFiles;
	iFiles++;
	pthread_mutex_unlock(&lock);
	return index;
}

void * doCreateFile(void * args) {
	unsigned long long index = 0;
	while ((index = getIndex()) < nbFiles) {
		char filename[width + 9];
		unsigned int dirIndex = index >> bshift;
		unsigned long long fileIndex = index & ((1 << bshift) - 1);
		snprintf(filename, width + 9, "d_%03x/f_%0*llx", dirIndex, width, fileIndex);

		int fd = open(filename, openOption, 0644);
		write(fd, buffer, 10240);
		close(fd);
	}
}

void * doComputeSize(void * args) {
	unsigned long long index = 0;
	while ((index = getIndex()) < nbFiles) {
		char filename[width + 9];
		unsigned int dirIndex = index >> bshift;
		unsigned long long fileIndex = index & ((1 << bshift) - 1);
		snprintf(filename, width + 9, "d_%03x/f_%0*llx", dirIndex, width, fileIndex);

		struct stat st;
		stat(filename, &st);

		pthread_mutex_lock(&lSize);
		totalSize += st.st_blocks;
		pthread_mutex_unlock(&lSize);
	}
}

void * doDisplay(void * args) {
	unsigned long long index = 0;

	struct timeval begin;
	gettimeofday(&begin, 0);

	sleep(2);

	for (;;) {
		short i;
		for (i = 0; i < 4; i++) {
			usleep(250000);

			pthread_mutex_lock(&lock);
			index = iFiles;
			pthread_mutex_unlock(&lock);

			if (index >= nbFiles)
				break;

			clean(0);
			report2("\rworking... eta: %E, %40b [%p]", index);
		}

		struct timeval end;
		gettimeofday(&end, 0);

		unsigned long long nbSecs = difftime(end.tv_sec, begin.tv_sec);
		if ((nbSecs & 0x3FFF) == 0) {
			struct tm now;
			localtime_r(&(end.tv_sec), &now);

			char buffer[32];
			strftime(buffer, 32, "%F %T", &now);

			clean(0);
			mkPrintf("At %s, ", buffer);
			report2("eta: %E [%p]\n", index);
		}
	}
}

void * doUnlink(void * args) {
	unsigned long long index = 0;
	while ((index = getIndex()) < nbFiles) {
		char filename[width + 9];
		unsigned int dirIndex = index >> bshift;
		unsigned long long fileIndex = index & ((1 << bshift) - 1);
		snprintf(filename, width + 9, "d_%03x/f_%0*llx", dirIndex, width, fileIndex);

		unlink(filename);
	}
}

void doJob(void * (*job)(void *), unsigned short nbJobs) {
	iFiles = 0;
	if (nbJobs > 1) {
		nbJobs--;

		pthread_t jobs[nbJobs];
		unsigned short i;
		for (i = 0; i < nbJobs; i++)
			pthread_create(jobs + i, 0, job, 0);

		pthread_t display;
		pthread_create(&display, 0, doDisplay, 0);

		job(0);

		for (i = 0; i < nbJobs; i++)
			pthread_join(jobs[i], 0);

		pthread_cancel(display);
		pthread_join(display, 0);
	} else {
		pthread_t display;
		pthread_create(&display, 0, doDisplay, 0);

		job(0);

		pthread_cancel(display);
		pthread_join(display, 0);
	}
	clean(1);
}

int main(int argc, char ** argv) {
	static int returnCode = 0;
	static char * wd = 0;
	static int dontDeleteFiles = 0;
	static unsigned short nbThread = 1, doPause = 0;

	static int option_index = 0;
	static struct option long_options[] = {
		{"dir",	    		0,	0,	'd'},
		{"dontDeleteFiles",	0,	0,  128},
		{"filesize",    	1,	0,	's'},
		{"help",    		0,	0,	'h'},
		{"nbFiles",	    	1,	0,	'f'},
		{"no-direct-io",    0,  0,  'D'},
		{"pause",           0,  0,  'p'},
		{"threads",	    	1,	0,	't'},
		{"version",		    0,	0,	'v'},

		{0, 0, 0, 0}
	};


	for (;;) {
		int c = getopt_long(argc, argv, "d:Df:hpst:V", long_options, &option_index);

		if (c == -1)
			break;

		switch (c) {
			case 'd':
				if (wd) {
					mkPrintf("Warning: using '%s' instead of '%s' as working directory\n", optarg, wd);
					free(wd);
				}
				wd = strdup(optarg);
				break;

			case 'D':
				openOption &= ~O_DIRECT;
				break;

			case 'f':
				if (sscanf(optarg, "%llu", &nbFiles) < 1) {
					mkPrintf("'%s' is not a valid argument for '-f' option\n", optarg);
					returnCode = 1;
					goto end;
				}
				break;

			case 'h': {
				char * ptr = strrchr(*argv, '/');
				if (!ptr)
					ptr = *argv;
				else
					ptr++;

				mkPrintf("%s, version: %s\n", ptr, MK1BFILE_VERSION);
				mkPrintf("   --dir,     -d : set working directory to\n");
				mkPrintf("                   default value : current working directory\n");
				mkPrintf("   --dontDeleteFiles : do not delete files\n");
				mkPrintf("   --help,    -h : show this and exit\n");
				mkPrintf("   --nbFiles, -f : set number of files that will be created and deleted\n");
				mkPrintf("                   default value : %lld files\n", NBFILES);
				mkPrintf("   --no-direct-io, -D : don't open with `O_DIRECT` flag\n");
				mkPrintf("   --pause,   -p : make a pause before deleting files\n");
				mkPrintf("   --threads, -t : set number of threads will be created\n");
				mkPrintf("                   default value : 1 thread\n");
				mkPrintf("                   should be between 1 and 64\n");
				mkPrintf("   --version, -V : show version and exit\n");
				goto end;
			}

			case 'p':
				doPause = 1;
				break;

			case 't':
				if (sscanf(optarg, "%llu", &nbThread) < 1) {
					mkPrintf("'%s' is not a valid argument for '-t' option\n", optarg);
					returnCode = 1;
					goto end;
				}
				if (nbThread < 1 || nbThread > 64) {
					mkPrintf("Number of thread should be between 1 and 64\n", optarg);
					returnCode = 1;
					goto end;
				}
				break;

			case 'V': {
				char * ptr = strrchr(*argv, '/');
				if (!ptr)
					ptr = *argv;
				else
					ptr++;

				mkPrintf("%s version: %s, build: %s %s\n", ptr, MK1BFILE_VERSION, __DATE__, __TIME__);
				return 0;

			}
			case 128:
				dontDeleteFiles = 1;
				break;
		}
	}

	bshift = nbFiles < 1024 ? 0 : floor(log((nbFiles - 1) >> 10) / log(2)) + 1;
	width = bshift > 0 ? floor(log((1 << bshift) - 1) / log(16)) + 1 : 1;
	setMax(nbFiles);

	if (wd && chdir(wd)) {
		mkPrintf("Filed to change working directory (%s) because '%m'\n", wd);
		free(wd);
		return 1;
	}


	static char buffer[64];
	char end = nbThread > 1 ? 's' : '\0';

	if (dontDeleteFiles)
		mkPrintf("mk1BFiles will create %lld files with %u thread%c\n", nbFiles, nbThread, end);
	else
		mkPrintf("mk1BFiles will create and delete %lld files with %u thread%c\n", nbFiles, nbThread, end);
	mkPrintf("Version: %s, build: %s %s\n\n", MK1BFILE_VERSION, __DATE__, __TIME__);

	getCurrentTime(buffer, 64);
	mkPrintf("Creating %lld files, started at %s...\n", nbFiles, buffer);
	restartTimer();
	static int i;
	for (i = 0; i < 1024; i++) {
		static char dirname[6];
		snprintf(dirname, 6, "d_%03x", i);

		if (access(dirname, R_OK))
			mkdir(dirname, 0766);
	}
	doJob(doCreateFile, nbThread);
	stopTimer();
	report("Done, time spent: %L\n\n");

	getCurrentTime(buffer, 64);
	mkPrintf("Doing `ls -R`, started at %s...\n", buffer);
	static const char * params[] = { 0, "-R", 0 };
	restartTimer();
	run("ls", params);
	stopTimer();
	report("Done, time spent: %L\n\n");

	getCurrentTime(buffer, 64);
	mkPrintf("Compute size used by %lld files, started at %s...\n", nbFiles, buffer);
	restartTimer();
	doJob(doComputeSize, nbThread);

	convertSize(buffer, 64, totalSize << 9);
	mkPrintf("Size used by files: %s\n", buffer);

	static struct stat st;
	static unsigned long long totalDirSize = 0;
	stat(".", &st);
	totalDirSize = st.st_blocks;

	for (i = 0; i < 1024; i++) {
		static char dirname[6];
		snprintf(dirname, 6, "d_%03x", i);

		stat(dirname, &st);
		totalDirSize += st.st_blocks;
	}
	convertSize(buffer, 64, totalDirSize << 9);
	mkPrintf("Size used by directory: %s\n", buffer);

	totalSize += totalDirSize;
	convertSize(buffer, 64, totalSize << 9);
	mkPrintf("Size used (total): %s\n", buffer);
	stopTimer();
	report("Done, time spent: %L\n\n");

	if (dontDeleteFiles)
		goto end;

	if (doPause)
		kill(getpid(), SIGSTOP);

	getCurrentTime(buffer, 64);
	mkPrintf("Deleting %lld files, started at %s...\n", nbFiles, buffer);
	restartTimer();
	doJob(doUnlink, nbThread);
	for (i = 0; i < 1024; i++) {
		static char dirname[6];
		snprintf(dirname, 6, "d_%03x", i);

		rmdir(dirname);
	}
	stopTimer();
	report("Done, time spent: %L\n");


end:
	if (wd)
		free(wd);

	return returnCode;
}

